local insert = table.insert
local validate_functions = {
  exists = function(input)
    return input and input ~= "", "%s must be provided"
  end,
  file_exists = function(input)
    return type(input) == "table" and input.filename ~= "" and input.content ~= "", "Missing file"
  end,
  min_length = function(input, len)
    return #tostring(input or "") >= len, "%s must be at least " .. tostring(len) .. " chars"
  end,
  max_length = function(input, len)
    return #tostring(input or "") <= len, "%s must be at most " .. tostring(len) .. " chars"
  end,
  is_integer = function(input)
    return tostring(input):match("^%d+$"), "%s must be an integer"
  end,
  is_color = (function()
    local hex = "[a-fA-f0-9]"
    local three = "^#" .. tostring(hex:rep(3)) .. "$"
    local six = "^#" .. tostring(hex:rep(6)) .. "$"
    return function(input)
      input = tostring(input)
      return input:match(three) or input:match(six), "%s must be a color"
    end
  end)(),
  equals = function(input, value)
    return input == value, "%s must match"
  end,
  one_of = function(input, ...)
    local choices = {
      ...
    }
    local _list_0 = choices
    for _index_0 = 1, #_list_0 do
      local choice = _list_0[_index_0]
      if input == choice then
        return true
      end
    end
    return false, "%s must be one of " .. tostring(table.concat(choices, ", "))
  end
}
local test_input
test_input = function(input, func, args)
  local fn = assert(validate_functions[func], "Missing validation function " .. tostring(func))
  if type(args) ~= "table" then
    args = {
      args
    }
  end
  return fn(input, unpack(args))
end
local validate
validate = function(object, validations)
  local errors = { }
  local _list_0 = validations
  for _index_0 = 1, #_list_0 do
    local _continue_0 = false
    repeat
      local v = _list_0[_index_0]
      local key = v[1]
      local error_msg = v[2]
      local input = object[key]
      if v.optional then
        if not (validate_functions.exists(input)) then
          _continue_0 = true
          break
        end
      end
      v.optional = nil
      for fn, args in pairs(v) do
        local _continue_1 = false
        repeat
          if not (type(fn) == "string") then
            _continue_1 = true
            break
          end
          local success, msg = test_input(input, fn, args)
          if not (success) then
            insert(errors, (error_msg or msg):format(key))
            break
          end
          _continue_1 = true
        until true
        if not _continue_1 then
          break
        end
      end
      _continue_0 = true
    until true
    if not _continue_0 then
      break
    end
  end
  return next(errors) and errors
end
local assert_valid
assert_valid = function(object, validations)
  local errors = validate(object, validations)
  if errors then
    return coroutine.yield("error", errors)
  end
end
if ... == "test" then
  require("moon")
  local o = {
    age = "",
    name = "abc",
    height = "12234234"
  }
  moon.p(validate(o, {
    {
      "age",
      exists = true
    },
    {
      "name",
      exists = true
    },
    {
      "rupture",
      exists = true,
      "CUSTOM MESSAGE COOL"
    }
  }))
  moon.p(validate(o, {
    {
      "name",
      exists = true,
      min_length = 4
    },
    {
      "age",
      min_length = 4
    },
    {
      "height",
      max_length = 5
    }
  }))
  moon.p(validate(o, {
    {
      "height",
      is_integer = true
    },
    {
      "name",
      is_integer = true
    },
    {
      "age",
      is_integer = true
    }
  }))
  moon.p(validate(o, {
    {
      "height",
      min_length = 4
    }
  }))
  moon.p(validate(o, {
    {
      "age",
      optional = true,
      max_length = 2
    },
    {
      "name",
      optional = true,
      max_length = 2
    }
  }))
  moon.p(validate(o, {
    {
      "name",
      one_of = {
        "cruise",
        "control"
      }
    }
  }))
  moon.p(validate(o, {
    {
      "name",
      one_of = {
        "bcd",
        "abc"
      }
    }
  }))
end
return {
  validate = validate,
  assert_valid = assert_valid,
  test_input = test_input,
  validate_functions = validate_functions
}
